#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_result module

Elasticsearch query result test class.
"""

__docformat__ = 'restructuredtext'

from unittest import TestCase

from pyams_elastic.result import ElasticResult, ElasticResultRecord


sample_record1 = {
    '_score': 0.85,
    '_id': 1234,
    '_type': 'Thing',
    '_source': {
        'name': 'Grue',
        'color': 'Dark'
    }
}


sample_record2 = {
    '_score': 0.62,
    '_id': 1249,
    '_type': 'Thing',
    '_source': {
        'name': 'Widget',
        'color': 'Red'
    }
}


sample_result = {
    '_shards': {
        'failed': 0,
        'successful': 2,
        'total': 2
    },
    'hits': {
        'hits': [
            sample_record1,
            sample_record2,
        ],
        'max_score': 0.85,
        'total': 2
    },
    'suggest': {
        'check1': [
            {
                'text': 'sdlr',
                'length': 4,
                'options': [
                    {'text': 'sldr',
                     'freq': 35,
                     'score': 0.75},
                    {'text': 'sale',
                     'freq': 94,
                     'score': 0.5},
                ],
                'offset': 0,
            }
        ]
    },
    'timed_out': False,
    'took': 1
}


class TestResult(TestCase):
    """Result test case"""

    def _make_result(self):  # pylint: disable=no-self-use
        """Create result"""
        return ElasticResult(sample_result)

    def test_result_repr(self):
        """Test result output"""
        result = self._make_result()
        self.assertIn('total:2', repr(result))


class TestResultRecord(TestCase):
    """Result record test case"""

    def _make_record(self):  # pylint: disable=no-self-use
        """Create record"""
        return ElasticResultRecord(sample_record1)

    def test_record_repr(self):
        """Test record output"""
        record = self._make_record()
        s = repr(record)  # pylint: disable=invalid-name
        self.assertIn('Thing', s)
        self.assertIn('1234', s)

    def test_record_getitem(self):
        """Test record item getter"""
        record = self._make_record()
        self.assertEqual(record['_type'], 'Thing')

    def test_record_attr_source(self):
        """Test record source attribute"""
        record = self._make_record()
        self.assertEqual(record.name, 'Grue')

    def test_record_attr_raw(self):
        """Test record raw attribute"""
        record = self._make_record()
        self.assertEqual(record._id, 1234)  # pylint: disable=protected-access

    def test_record_attr_nonexistent(self):
        """Test record missing attribute"""
        record = self._make_record()
        with self.assertRaises(AttributeError):
            record.nonexistent  # pylint: disable=pointless-statement

    def test_record_contains(self):
        """Test record container"""
        record = self._make_record()
        self.assertIn('_score', record)
        self.assertNotIn('foo', record)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.task.zmi module

This module provides management components for Elasticsearch tasks.
"""

from zope.interface import Interface, alsoProvides, implementer

from pyams_elastic.client import ElasticClientInfo
from pyams_elastic.interfaces import IElasticClientInfo
from pyams_elastic.task.interfaces import IElasticTask, IElasticTaskInfo
from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.group import GroupManager
from pyams_form.interfaces import IObjectFactory
from pyams_form.interfaces.form import IForm, IInnerTabForm
from pyams_form.interfaces.widget import IObjectWidget
from pyams_form.subform import InnerAddForm, InnerEditForm
from pyams_layer.interfaces import IFormLayer, IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, MANAGE_TASKS_PERMISSION
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_scheduler.zmi.task import BaseTaskAddForm, BaseTaskEditForm
from pyams_skin.viewlet.menu import MenuItem
from pyams_utils.adapter import adapter_config
from pyams_utils.factory import get_interface_name
from pyams_utils.interfaces.data import IObjectData
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager


__docformat__ = 'restructuredtext'

from pyams_elastic import _  # pylint: disable=ungrouped-imports


@adapter_config(name=get_interface_name(IElasticClientInfo),
                required=(Interface, IFormLayer, IForm, IObjectWidget),
                provides=IObjectFactory)
def elastic_connection_factory(*args, **kwargs):  # pylint: disable=unused-argument
    """Elasticsearch connection object factory"""
    return ElasticClientInfo


class IElasticTaskForm(IForm):
    """Elasticsearch task form marker interface"""


@implementer(IElasticTaskForm)
class ElasticTaskFormInfo(GroupManager):
    """Elasticsearch task form info"""

    title = _("Elasticsearch task settings")
    fields = Fields(IElasticTaskInfo)

    def update_widgets(self, prefix=None):
        """Widgets update method"""
        super().update_widgets(prefix)  # pylint: disable=no-member
        query = self.widgets.get('query')  # pylint: disable=no-member
        if query is not None:
            query.add_class('height-100')
            query.widget_css_class = 'editor height-300px'
            query.object_data = {
                'ams-filename': 'query.json'
            }
            alsoProvides(query, IObjectData)
        fields = self.widgets.get('log_fields')  # pylint: disable=no-member
        if fields is not None:
            fields.rows = 5


@viewlet_config(name='add-elastic-task.menu',
                context=IScheduler, layer=IAdminLayer, view=SchedulerTasksTable,
                manager=IContextAddingsViewletManager, weight=110,
                permission=MANAGE_TASKS_PERMISSION)
class ElasticTaskAddMenu(MenuItem):
    """Elasticsearch task add menu"""

    label = _("Add Elasticsearch query...")
    href = 'add-elastic-task.html'
    modal_target = True


@ajax_form_config(name='add-elastic-task.html', context=IScheduler, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class ElasticTaskAddForm(BaseTaskAddForm):
    """Elasticsearch task add form"""

    content_factory = IElasticTask


@adapter_config(name='elastic-task-info.form',
                required=(IScheduler, IAdminLayer, ElasticTaskAddForm),
                provides=IInnerTabForm)
class ElasticTaskAddFormInfo(ElasticTaskFormInfo, InnerAddForm):
    """Elasticsearch task add form info"""


@ajax_form_config(name='properties.html', context=IElasticTask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class ElasticTaskEditForm(BaseTaskEditForm):
    """Elasticsearch task edit form"""


@adapter_config(name='elastic-task-info.form',
                required=(IElasticTask, IAdminLayer, ElasticTaskEditForm),
                provides=IInnerTabForm)
class ElasticTaskEditFormInfo(ElasticTaskFormInfo, InnerEditForm):
    """Elasticsearch task edit form info"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_pyramid module

Pyramid integration tests.
"""

__docformat__ = 'restructuredtext'

from unittest import TestCase
from urllib.parse import urlencode

from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPFound
from sqlalchemy import Column, types
from sqlalchemy.ext.declarative import declarative_base
from webtest import TestApp

from pyams_elastic.include import get_client
from pyams_elastic.mixin import ESKeyword, ESMapping, ESText, ElasticMixin


Base = declarative_base()


class Todo(Base, ElasticMixin):

    __tablename__ = 'todos'

    id = Column(types.Integer, primary_key=True)
    description = Column(types.Unicode(40))

    @classmethod
    def elastic_mapping(cls):
        return ESMapping(
            # analyzer='content',
            properties=ESMapping(
                ESKeyword('document_type'),
                ESText('description')))


def index_view(request):
    es_client = get_client(request)
    es_client.refresh()
    result = es_client.query(Todo).execute()
    return {
        'todos': [rec.to_dict() for rec in result],
        'count': result.total.value
    }


def add_view(request):
    es_client = get_client(request)
    for s in request.params['description'].split(', '):
        todo = Todo(description=s)
        es_client.index_object(todo)
    if request.params.get('fail_after_index'):
        raise RuntimeError('fail!')
    return HTTPFound(location=request.route_url('index'))


def make_app():
    settings = {
        'pyams_elastic.index': 'pyams_elastic_tests_app',
        'pyams_elastic.servers': ['elasticsearch:9200'],
    }
    config = Configurator(settings=settings)
    config.include('pyramid_tm')
    config.include('cornice')
    config.include('pyams_utils')
    config.include('pyams_elastic')

    config.add_route('index', '/')
    config.add_view(index_view, route_name='index', renderer='json')

    config.add_route('add', '/add')
    config.add_view(add_view, route_name='add')

    es_client = get_client(config)
    es_client.ensure_index(recreate=True)

    sample = Todo(description='Example to-do item')
    es_client.index_object(sample, immediate=True)

    return config.make_wsgi_app()


class PyramidTestLayer:

    @classmethod
    def setUp(cls):
        cls.app = TestApp(make_app())

    @classmethod
    def tearDown(cls):
        client = get_client(cls.app.app)
        client.delete_index()


class TestPyramid(TestCase):

    layer = PyramidTestLayer

    @classmethod
    def setUpClass(cls):
        cls.layer.setUp()

    @classmethod
    def tearDownClass(cls):
        cls.layer.tearDown()

    @property
    def app(self):
        return self.layer.app

    def test_index(self):
        resp = self.app.get('/')
        resp.mustcontain('Example')

    def test_add_successful(self):
        params = urlencode({
            'description': 'Zygomorphic',
        })
        self.app.get('/add?' + params, status=302)
        # Check that new todo is now in the index.
        resp = self.app.get('/')
        resp.mustcontain('Zygomorphic')

    def test_add_fail(self):
        params = urlencode({
            'description': 'Nucleoplasm',
            'fail_after_index': True,
        })
        with self.assertRaises(RuntimeError):
            self.app.get('/add?' + params)

        resp = self.app.get('/')
        # Check that new todo is *not* in the index.
        self.assertNotIn('Nucleoplasm', resp.body.decode('utf8'))

    def test_add_alternate(self):
        params = urlencode({
            'description': 'Banana',
        })
        self.app.get('/add?' + params, status=302)
        resp = self.app.get('/')
        resp.mustcontain('Banana')

        params = urlencode({
            'description': 'Apple',
            'fail_after_index': 1,
        })
        with self.assertRaises(RuntimeError):
            self.app.get('/add?' + params)
        resp = self.app.get('/')
        self.assertNotIn('Apple', resp.body.decode('utf8'))

        params = urlencode({
            'description': 'Kiwi, Pineapple, Cherry',
        })
        self.app.get('/add?' + params, status=302)
        resp = self.app.get('/')
        resp.mustcontain('Kiwi')



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_transaction module

This module is used to test transaction integration.
"""

__docformat__ = 'restructuredtext'

from unittest import TestCase

import transaction
from sqlalchemy import Column, types
from sqlalchemy.ext.declarative import declarative_base

from pyams_elastic.client import ElasticClient
from pyams_elastic.mixin import ESKeyword, ESMapping, ESText, ElasticMixin


Base = declarative_base()


class Todo(Base, ElasticMixin):

    __tablename__ = 'todos'

    id = Column(types.Integer, primary_key=True)
    description = Column(types.Unicode(40))

    @classmethod
    def elastic_mapping(cls):
        return ESMapping(
            # analyzer='content',
            properties=ESMapping(
                ESKeyword('document_type'),
                ESText('description')))


class TestClient(TestCase):

    def setUp(self):
        self.client = ElasticClient(servers=['elasticsearch:9200'],
                                    index='pyams_elastic_tests_txn',
                                    use_transaction=True)
        self.client.ensure_index(recreate=True)
        self.client.ensure_all_mappings(Base, recreate=True)

    def tearDown(self):
        self.client.delete_index()
        self.client.close()

    def test_index_and_delete_document(self):
        todo = Todo(id=42, description='Finish exhaustive test suite')

        with transaction.manager:
            self.client.index_object(todo)
        self.client.flush(force=True)
        self.client.refresh()

        # Search for this document and make sure it exists.
        q = self.client.query(Todo, query='exhaustive')
        result = q.execute()
        todos = [doc.description for doc in result]
        self.assertIn('Finish exhaustive test suite', todos)

        with transaction.manager:
            self.client.delete_object(todo)

        self.client.flush(force=True)
        self.client.refresh()

        # Search for this document and make sure it DOES NOT exist.
        q = self.client.query(Todo, query='exhaustive')
        result = q.execute()
        todos = [doc.description for doc in result]
        self.assertNotIn('Finish exhaustive test suite', todos)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS elastic.include module

This module is used for Pyramid integration
"""

__docformat__ = 'restructuredtext'

from pyramid.settings import asbool

from pyams_elastic.client import ElasticClient
from pyams_elastic.interfaces import IElasticClient


def client_from_config(settings, prefix='pyams_elastic.'):
    """
    Instantiate and configure an Elasticsearch from settings.

    In typical Pyramid usage, you shouldn't use this directly: instead, just
    include ``pyams_elastic`` and use the :py:func:`get_client` function to get
    access to the shared :py:class:`.client.ElasticClient` instance.
    """
    return ElasticClient(
        servers=settings.get(prefix + 'servers', ['elasticsearch:9200']),
        index=settings[prefix + 'index'],
        timeout=settings.get(prefix + 'timeout', 10.0),
        timeout_retries=int(settings.get(prefix + 'timeout_retry', 0)),
        use_transaction=asbool(settings.get(prefix + 'use_transaction', True)),
        disable_indexing=settings.get(prefix + 'disable_indexing', False))


def get_client(request):
    """
    Get the registered Elasticsearch client. The supplied argument can be
    either a ``Request`` instance or a ``Registry``.
    """
    registry = request.registry
    return registry.queryUtility(IElasticClient)


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_elastic:locales')

    # add request methods
    config.add_request_method(get_client, 'elastic_client', reify=True)

    # initialize Elasticsearch client
    registry = config.registry
    settings = registry.settings
    if settings.get('pyams_elastic.index'):
        client = client_from_config(settings)
        if asbool(settings.get('pyams_elastic.ensure_index_on_start')):
            client.ensure_index()
        registry.registerUtility(client, IElasticClient)

    # package scan
    ignored = []
    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        ignored.append('pyams_elastic.zmi')
        ignored.append('pyams_elastic.task.zmi')

    try:
        import pyams_scheduler  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        ignored.append('pyams_elastic.task')

    config.scan(ignore=ignored)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.client module

This module defines the main Elasticsearch client class.
"""

__docformat__ = 'restructuredtext'

import logging
from functools import wraps
from pprint import pformat

import transaction as zope_transaction
from elasticsearch import Elasticsearch, NotFoundError
from persistent import Persistent
from transaction.interfaces import ISavepointDataManager
from zope.component import getAdapters
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_elastic.docdict import DotDict
from pyams_elastic.interfaces import IElasticClient, IElasticClientInfo, IElasticMapping, \
    IElasticMappingExtension
from pyams_elastic.query import ElasticQuery
from pyams_utils.factory import factory_config


LOGGER = logging.getLogger('PyAMS (elastic)')


ANALYZER_SETTINGS = {
    "analysis": {
        "filter": {
            "snowball": {
                "type": "snowball",
                "language": "English"
            },
        },

        "analyzer": {
            "lowercase": {
                "type": "custom",
                "tokenizer": "standard",
                "filter": ["lowercase"]
            },

            "email": {
                "type": "custom",
                "tokenizer": "uax_url_email",
                "filter": ["lowercase"]
            },

            "content": {
                "type": "custom",
                "tokenizer": "standard",
                "char_filter": ["html_strip"],
                "filter": ["lowercase", "stop", "snowball"]
            }
        }
    }
}


CREATE_INDEX_SETTINGS = ANALYZER_SETTINGS.copy()
CREATE_INDEX_SETTINGS.update({
    "index": {
        "number_of_shards": 2,
        "number_of_replicas": 0
    },
})

STATUS_ACTIVE = 'active'
STATUS_CHANGED = 'changed'


_CLIENT_STATE = {}


class ElasticSavepoint:
    """Elasticsearch savepoint"""

    def __init__(self, dm):
        self.dm = dm  # pylint: disable=invalid-name
        self.saved = dm.client.uncommitted.copy()

    def rollback(self):
        """Savepoint rollback"""
        self.dm.client.uncommitted = self.saved.copy()


@implementer(ISavepointDataManager)
class ElasticDataManager:
    """Elasticsearch data manager"""

    def __init__(self, client, transaction_manager):
        self.client = client
        self.transaction_manager = transaction_manager
        t = transaction_manager.get()  # pylint: disable=invalid-name
        t.join(self)
        _CLIENT_STATE[id(client)] = STATUS_ACTIVE
        self._reset()

    def _reset(self):
        """Data manager reset"""
        LOGGER.debug('_reset(%s)', self)
        self.client.uncommitted = []

    def _finish(self):
        """Data manager finish"""
        LOGGER.debug('_finish(%s)', self)
        client = self.client
        del _CLIENT_STATE[id(client)]

    def abort(self, transaction):  # pylint: disable=unused-argument
        """Transaction abort"""
        LOGGER.debug('abort(%s)', self)
        self._reset()
        self._finish()

    def tpc_begin(self, transaction):  # pylint: disable=unused-argument
        """Begin two-phases commit"""
        LOGGER.debug('tpc_begin(%s)', self)

    def commit(self, transaction):  # pylint: disable=unused-argument
        """Transaction commit"""
        LOGGER.debug('commit(%s)', self)

    def tpc_vote(self, transaction):  # pylint: disable=unused-argument
        """Two-phases commit vote"""
        LOGGER.debug('tpc_vote(%s)', self)
        # XXX: Ideally, we'd try to check the uncommitted queue and make sure
        # everything looked ok. Not sure how we can do that, though.

    def tpc_finish(self, transaction):  # pylint: disable=unused-argument
        """Two-phases commit finish"""
        # Actually persist the uncommitted queue.
        LOGGER.debug('tpc_finish(%s)', self)
        LOGGER.warning("running: %r", self.client.uncommitted)
        for cmd, args, kwargs in self.client.uncommitted:
            kwargs['immediate'] = True
            getattr(self.client, cmd)(*args, **kwargs)
        self._reset()
        self._finish()

    def tpc_abort(self, transaction):  # pylint: disable=unused-argument
        """Two-phases commit abort"""
        LOGGER.debug('tpc_abort()')
        self._reset()
        self._finish()

    def sortKey(self):  # pylint: disable=invalid-name
        """Data manager sort key getter"""
        # NOTE: Ideally, we want this to sort *after* database-oriented data
        # managers, like the SQLAlchemy one. The double tilde should get us
        # to the end.
        return '~~elasticsearch' + str(id(self))

    def savepoint(self):
        """Savepoint getter"""
        return ElasticSavepoint(self)


def join_transaction(client, transaction_manager):
    """Join current transaction"""
    client_id = id(client)
    existing_state = _CLIENT_STATE.get(client_id, None)
    if existing_state is None:
        LOGGER.warning('client %s not found, setting up new data manager',
                       client_id)
        ElasticDataManager(client, transaction_manager)
    else:
        LOGGER.warning('client %s found, using existing data manager',
                       client_id)
        _CLIENT_STATE[client_id] = STATUS_CHANGED


def transactional(f):  # pylint: disable=invalid-name
    """Transactional functions wrapper"""

    @wraps(f)
    def transactional_inner(client, *args, **kwargs):
        """Inner transaction wrapper"""
        immediate = kwargs.pop('immediate', None)
        if client.use_transaction:
            if immediate:
                return f(client, *args, **kwargs)
            LOGGER.debug('enqueueing action: %s: %r, %r', f.__name__, args, kwargs)
            join_transaction(client, client.transaction_manager)
            client.uncommitted.append((f.__name__, args, kwargs))
            return None
        return f(client, *args, **kwargs)
    return transactional_inner


@factory_config(IElasticClientInfo)
@implementer(IElasticClientInfo)
class ElasticClientInfo(Persistent):
    """Elasticsearch client connection info"""

    servers = FieldProperty(IElasticClientInfo['servers'])
    use_ssl = FieldProperty(IElasticClientInfo['use_ssl'])
    verify_certs = FieldProperty(IElasticClientInfo['verify_certs'])
    index = FieldProperty(IElasticClientInfo['index'])
    timeout = FieldProperty(IElasticClientInfo['timeout'])
    timeout_retries = FieldProperty(IElasticClientInfo['timeout_retries'])

    def __init__(self, data=None):  # pylint: disable=unused-argument
        super().__init__()

    def open(self):
        """Open Elasticsearch client"""
        return Elasticsearch(self.servers,  # pylint: disable=invalid-name
                             use_ssl=self.use_ssl,
                             verify_certs=self.verify_certs,
                             timeout=self.timeout,
                             retry_on_timeout=self.timeout_retries > 0,
                             max_retries=self.timeout_retries)


@implementer(IElasticClient)
class ElasticClient:
    """
    A handle for interacting with the Elasticsearch backend.
    """

    def __init__(self, servers=None, index=None, using=None,
                 auth=None,
                 use_ssl=False,
                 verify_certs=True,
                 timeout=10.0,
                 timeout_retries=0,
                 disable_indexing=False,
                 use_transaction=True,
                 transaction_manager=zope_transaction.manager):
        # pylint: disable=too-many-arguments,unused-argument
        assert servers or using, "You must provide servers or connection info!"
        self.disable_indexing = disable_indexing
        self.use_transaction = use_transaction
        self.transaction_manager = transaction_manager
        if using is not None:
            self.index = using.index
            self.es = using.open()  # pylint: disable=invalid-name
        else:
            self.index = index
            self.es = Elasticsearch(servers,  # pylint: disable=invalid-name
                                    auth=auth,
                                    use_ssl=use_ssl,
                                    verify_certs=verify_certs,
                                    timeout=timeout,
                                    retry_on_timeout=timeout_retries > 0,
                                    max_retries=timeout_retries)

    def close(self):
        """Close Elasticsearch client"""
        self.es.close()

    def __enter__(self):
        return self.es

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.es.close()

    def ensure_index(self, recreate=False, settings=None):
        """
        Ensure that the index exists on the ES server, and has up-to-date
        settings.
        """
        exists = self.es.indices.exists(self.index)
        if recreate or not exists:
            if exists:
                self.es.indices.delete(self.index)
            self.es.indices.create(self.index,
                                   body=dict(settings=settings or CREATE_INDEX_SETTINGS))

    def delete_index(self):
        """
        Delete the index on the ES server.
        """
        self.es.indices.delete(self.index)

    def get_mappings(self):  # pylint: disable=unused-argument
        """
        Return the object mappings currently used by ES.
        """
        raw = self.es.indices.get_mapping(index=self.index)
        return raw[self.index]['mappings']

    def ensure_mapping(self, cls, recreate=False):
        """
        Put an explicit mapping for the given class if it doesn't already
        exist.
        """
        doc_mapping = cls.elastic_mapping()
        doc_mapping = dict(doc_mapping)

        LOGGER.debug('Putting mapping: \n%s', pformat(doc_mapping))
        mappings = self.get_mappings()
        if (not mappings) or recreate:
            self.es.indices.put_mapping(index=self.index,
                                        body=doc_mapping)

    def ensure_all_mappings(self, base_class, recreate=False):
        """
        Initialize explicit mappings for all subclasses of the specified
        SQLAlchemy declarative base class.
        """
        doc_mapping = self.get_mappings()
        if (not doc_mapping) or recreate:
            doc_mapping = {}
            for cls in base_class._decl_class_registry.values():  # pylint: disable=protected-access
                if not IElasticMapping.providedBy(cls):
                    continue
                cls_mapping = dict(cls.elastic_mapping())
                if cls_mapping:
                    for _name, extension in getAdapters((cls,), IElasticMappingExtension):
                        cls_mapping.update(extension.elastic_mapping())
                    doc_mapping.update(cls_mapping)
            LOGGER.debug('Putting mapping: \n%s', pformat(doc_mapping))
            self.es.indices.put_mapping(index=self.index,
                                        body=doc_mapping)

    def index_objects(self, objects):
        """
        Add multiple objects to the index.
        """
        for obj in objects:
            self.index_object(obj)

    def index_object(self, obj, **kw):
        """
        Add or update the indexed document for an object.
        """
        doc = obj.elastic_document()
        doc_id = doc.pop("_id")

        LOGGER.debug('Indexing object:\n%s', pformat(doc))
        LOGGER.debug('ID is %r', doc_id)

        self.index_document(id=doc_id, doc=doc, **kw)

    @transactional
    def index_document(self, id, doc):  # pylint: disable=invalid-name,redefined-builtin
        """
        Add or update the indexed document from a raw document source (not an
        object).
        """
        if self.disable_indexing:
            return

        kwargs = dict(index=self.index, body=doc, id=id)
        if '__pipeline__' in doc:
            kwargs['pipeline'] = doc.pop('__pipeline__')
        self.es.index(**kwargs)

    def delete_object(self, obj, safe=False, **kw):
        """
        Delete the indexed document for an object.
        """
        doc = obj.elastic_document()
        doc_id = doc.pop("_id")

        self.delete_document(id=doc_id, safe=safe, **kw)

    @transactional
    def delete_document(self, id, safe=False):  # pylint: disable=invalid-name,redefined-builtin
        """
        Delete the indexed document based on a raw document source (not an
        object).
        """
        if self.disable_indexing:
            return

        kwargs = dict(index=self.index, id=id)
        try:
            self.es.delete(**kwargs)
        except NotFoundError:
            if not safe:
                raise

    def flush(self, force=True):
        """
        Flush indices data
        """
        self.es.indices.flush(index=self.index, force=force)  # pylint: disable=unexpected-keyword-arg

    def get(self, obj):
        """
        Retrieve the ES source document for a given object or (document type,
        id) pair.
        """
        if isinstance(obj, (list, tuple)):
            _doc_type, doc_id = obj
        else:
            doc_id = obj.id

        kwargs = dict(index=self.index, id=doc_id)
        r = self.es.get(**kwargs)  # pylint: disable=invalid-name
        return DotDict(r['_source'])

    def refresh(self):
        """Refresh the ES index."""
        self.es.indices.refresh(index=self.index)

    def query(self, *classes, **kw):
        """
        Return an ElasticQuery against the specified class.
        """
        cls = kw.pop('cls', ElasticQuery)
        return cls(client=self, classes=classes, **kw)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.query module

This module defines Elasticsearch query class.
"""

__docformat__ = 'restructuredtext'

import inspect
from functools import wraps

from elasticsearch_dsl import A, Q, Search

from pyams_elastic.mixin import ElasticMixin
from pyams_elastic.result import ElasticResult


ARBITRARILY_LARGE_SIZE = 10000


def filters(f):  # pylint: disable=invalid-name
    """Query filters appender

    A convenience decorator to wrap query methods that are adding filters. To
    use, simply make a method that returns a filter dict in elasticsearch's
    JSON object format.
    """
    @wraps(f)
    def wrapped(self, *args, **kwargs):
        val = f(self, *args, **kwargs)
        if self.query is None:
            self.query = val
        else:
            self.query &= val
        return self
    return wrapped


def sorting(f):  # pylint: disable=invalid-name
    """Update current search sort order"""
    @wraps(f)
    def wrapped(self, *args, **kwargs):
        val = f(self, *args, **kwargs)
        self.sorts.append(val)
        return self
    return wrapped


def bucket(f):  # pylint: disable=invalid-name
    """Update query aggregates"""
    @wraps(f)
    def wrapped(self, name, *args, **kwargs):
        val = f(self, name, *args, **kwargs)
        self.dsl.aggs.bucket(name, val)
        return self
    return wrapped


class ElasticQuery:
    """
    Wrapper around Elasticsearch-DSL Search object
    """

    def __init__(self, client, classes=None, query=None):
        self.client = client
        self.dsl = Search(using=client.es,
                          index=client.index)
        if isinstance(query, str):
            self.query = Q('simple_query_string', **{'query': query})
        else:
            self.query = query
        if classes:
            docs_types = [
                cls.get_document_type()
                if inspect.isclass(cls) and issubclass(cls, ElasticMixin) else cls
                for cls in classes
            ]
            docs_types_filter = \
                Q('terms', **{'document_type': docs_types}) | \
                Q('terms', **{'document_type.keyword': docs_types})
            self.filter(docs_types_filter)
        self.sorts = []
        self.aggs = []
        self._start = None
        self._size = None

    @filters
    def ids(self, values):  # pylint: disable=no-self-use
        """Add filter documents on IDs"""
        if not isinstance(values, (list, tuple, set)):
            values = {values}
        return Q('ids', **{'values': list(map(str, values))})

    @filters
    def match_all_query(self):  # pylint: disable=no-self-use
        """Static method to return a filter which will match everything"""
        return Q('match_all')

    @filters
    def text_query(self, phrase, fields=None):  # pylint: disable=no-self-use
        """Static method to return a filter matching a text query"""
        query = {'query': phrase}
        if fields:
            query['fields'] = fields
        return Q('multi_match', **query)

    @filters
    def filter(self, query):  # pylint: disable=no-self-use,redefined-builtin
        """Add custom query to filters"""
        return query

    @filters
    def match(self, term, value, **kwargs):  # pylint: disable=no-self-use
        """Add match filter"""
        query = {'query': value}
        query.update(kwargs)
        return Q('match', **{term: query})

    @filters
    def filter_term(self, term, value, boost=1.0):  # pylint: disable=no-self-use
        """Add term filter"""
        return Q('term', **{term: {'value': value, 'boost': boost}})

    @filters
    def filter_terms(self, term, values, boost=1.0):  # pylint: disable=no-self-use
        """Add terms filters"""
        return Q('terms', **{term: values, 'boost': boost})

    @filters
    def filter_value_upper(self, term, upper):  # pylint: disable=no-self-use
        """Filter documents for which term is lower or equal to *upper*"""
        return Q('range', **{term: {'to': upper, 'include_upper': True}})

    @filters
    def filter_value_lower(self, term, lower):  # pylint: disable=no-self-use
        """Filter documents for which term is higher or equal to *lower*"""
        return Q('range', **{term: {'from': lower, 'include_lower': True}})

    @sorting
    def order_by(self, key, desc=False, **kwargs):  # pylint: disable=no-self-use
        """Sort results on given fields in ascending order, unless if *desc* is True"""
        order = {'order': 'desc' if desc else 'asc'}
        order.update(kwargs)
        return {key: order}

    def offset(self, n):  # pylint: disable=invalid-name
        """Start results at document ``n``"""
        if self._start is not None:
            raise ValueError('This query already has an offset applied.')
        self._start = n
        return self
    start = offset

    def limit(self, n):  # pylint: disable=invalid-name
        """Limit results at document ``n``"""
        if self._size is not None:
            raise ValueError('This query already has a limit applied.')
        self._size = n
        return self
    size = limit

    @bucket
    def add_term_aggregate(self, name, field, **kwargs):
        # pylint: disable=no-self-use,unused-argument
        """Add simple term aggregate"""
        return A('terms', field=field, **kwargs)

    @bucket
    def add_range_aggregate(self, name, field, ranges, **kwargs):
        # pylint: disable=no-self-use,unused-argument
        """Add range aggregate"""
        return A('range', field=field, ranges=ranges, **kwargs)

    def add_term_suggester(self, name, field, text):  # pylint: disable=no-self-use
        """Add term suggester"""
        self.dsl = self.dsl.suggest(name, text, term={'field': field})
        return self

    def search(self, start=0, size=ARBITRARILY_LARGE_SIZE, fields=None):
        """Start search"""
        if not self.query:
            self.query = self.match_all_query()
        query = self.dsl.query(self.query)

        if self.sorts:
            query = query.sort(*self.sorts)

        q_start = self._start or 0
        q_size = self._size or ARBITRARILY_LARGE_SIZE

        if start is not None:
            q_start = q_start + start
        if size is not None:
            q_size = max(0,
                         size if q_size is None else min(size, q_size - q_start))

        if q_start:
            q_size = min(q_size, ARBITRARILY_LARGE_SIZE - q_start)

        if q_start:
            query = query.extra(from_=q_start)
        if q_size:
            query = query.extra(size=q_size)

        if fields:
            query = query.source(fields)

        return query.execute()

    def execute(self, start=None, size=None, fields=None):
        """Execute this query and return a result set."""
        return ElasticResult(self.search(start=start, size=size, fields=fields))

    def count(self):
        """Return search total results"""
        res = self.search(size=0)
        return res.hits.total.value



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""
Generic test case for pyams_elastic doctests
"""

__docformat__ = 'restructuredtext'

import doctest
import os
import unittest

from pyams_elastic.tests import get_package_dir


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def doc_suite(test_dir, setUp=None, tearDown=None, globs=None):  # pylint: disable=invalid-name
    """Returns a test suite, based on doctests found in /doctests"""
    suite = []
    if globs is None:
        globs = globals()

    flags = (doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE |
             doctest.REPORT_ONLY_FIRST_FAILURE)

    package_dir = get_package_dir(test_dir)
    doctest_dir = os.path.join(package_dir, 'doctests')

    # filtering files on extension
    docs = [os.path.join(doctest_dir, doc) for doc in
            os.listdir(doctest_dir) if doc.endswith('.txt') or doc.endswith('.rst')]

    for test in docs:
        suite.append(doctest.DocFileSuite(test, optionflags=flags,
                                          globs=globs, setUp=setUp,
                                          tearDown=tearDown,
                                          module_relative=False))

    return unittest.TestSuite(suite)


def test_suite():
    """Returns the test suite"""
    return doc_suite(CURRENT_DIR)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS elastic.interfaces module

This module defines mail package interfaces.
"""

from zope.interface import Interface
from zope.schema import Bool, Float, Int, TextLine

from pyams_utils.schema import TextLineListField

from pyams_elastic import _


class IElasticClientInfo(Interface):
    """Elasticsearch client information interface"""

    servers = TextLineListField(title=_("Servers"),
                                description=_("Newline separated list of Elasticsearch servers "
                                              "URLs"),
                                required=True,
                                default=['elasticsearch:9200'])

    use_ssl = Bool(title=_("Use SSL?"),
                   description=_("If 'yes', use SSL for HTTP connection"),
                   required=True,
                   default=False)

    verify_certs = Bool(title=_("Verify certificates?"),
                        description=_("If 'no', SSL certificates will not be verified"),
                        required=True,
                        default=True)

    index = TextLine(title=_("Index name"),
                     description=_("Elasticsearch index name or pattern"),
                     required=True)

    timeout = Float(title=_("Timeout"),
                    description=_("Request timeout, in seconds"),
                    required=True,
                    default=10.0)

    timeout_retries = Int(title=_("Timeout retries"),
                          description=_("You can define the number of retries which can be done "
                                        "if a timeout occurs; setting this to 0 disable retries"),
                          required=False,
                          default=0)


class IElasticClient(IElasticClientInfo):
    """Elasticsearch client marker interface"""


class IElasticMapping(Interface):
    """Elasticsearch mapping interface"""

    def elastic_mapping(self):
        """Elasticsearch mapping getter

        This must be defined as a class method!
        """


class IElasticMappingExtension(IElasticMapping):
    """Elasticsearch mapping extension interface

    This interface defines an extension to an existing mapping interface.
    """



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_client module

This module provides a few unit tests about Elasticsearch client.
"""

from unittest import TestCase

from pyams_elastic.client import ElasticClient, ElasticClientInfo

__docformat__ = 'restructuredtext'


class TestClient(TestCase):

    def test_client_without_argument(self):
        with self.assertRaises(AssertionError) as context:
            client = ElasticClient()
        self.assertIn("You must provide servers or connection info!", str(context.exception))

    def test_client_with_servers(self):
        client = ElasticClient(servers=['elasticsearch:9200'],
                               index='pyams_elastic_tests',
                               use_transaction=False)
        with client as es:
            ping = es.ping(pretty=True)
            self.assertTrue(ping)
            info = es.info()
            self.assertIn('cluster_name', info)

    def test_client_with_info(self):
        info = ElasticClientInfo()
        info.servers = ['elasticsearch:9200']
        info.index = 'pyams_elastic_tests'
        client = ElasticClient(using=info,
                               use_transaction=False)
        with client as es:
            ping = es.ping(pretty=True)
            self.assertTrue(ping)
            info = es.info()
            self.assertIn('cluster_name', info)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.result module

This module defines Elasticsearch results management classes.
"""

__docformat__ = 'restructuredtext'

from pyams_elastic.docdict import DotDict


class ElasticResultRecord:
    """
    Wrapper for an Elasticsearch result record. Provides access to the indexed
    document, ES result data (like score), and the mapped object.
    """
    def __init__(self, raw):
        self.raw = DotDict(raw)

    def __repr__(self):
        return '<{} score:{} id:{} type:{}>'.format(self.__class__.__name__,
                                                    getattr(self, '_score', '-'),
                                                    self._id,
                                                    self._type)

    def __getitem__(self, key):
        return self.raw[key]

    def __contains__(self, key):
        return key in self.raw

    def __getattr__(self, key):
        source = self.raw.get('_source', {})
        fields = self.raw.get('fields', {})
        if key in source:
            return source[key]
        if key in fields:
            return fields[key]
        if key in self.raw:
            return self.raw[key]
        raise AttributeError('{!r} object has no attribute {!r}'.format(self.__class__.__name__,
                                                                        key))


class ElasticResult:
    """
    Wrapper for an Elasticsearch result set. Provides access to the documents,
    result aggregate data (like total count), facets and aggregations.

    Iterate over this object to yield document records, which are instances of
    :py:class:`ElasticResultRecord`.
    """
    def __init__(self, raw):
        self.raw = raw

    def __iter__(self):
        return iter(self.raw)

    def __repr__(self):
        return '<%s total:%s>' % (self.__class__.__name__, self.total)

    @property
    def total(self):
        """
        Return the total number of docs which would have been matched by this
        query. Note that this is not necessarily the same as the number of
        document result records associated with this object, because the query
        may have a start / size applied.
        """
        return self.raw['hits']['total']

    @property
    def facets(self):
        """
        Return the facets returned by this search query.
        """
        return self.raw.facets

    @property
    def aggregations(self):
        """
        Return aggregations returned by this search query
        """
        return self.raw.aggregations

    @property
    def suggests(self):
        """
        Return suggestions returned by this search query
        """
        return self.raw.suggest



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_mapping module

Small module used to test mappings.
"""

__docformat__ = 'restructuredtext'

from unittest import TestCase

from pyramid.config import Configurator
from sqlalchemy import Column, types

from pyams_elastic.interfaces import IElasticMappingExtension
from pyams_elastic.mixin import ESField, ESMapping, ESObject, ESText, ElasticMixin
from pyams_elastic.tests.data import Base, Movie
from pyams_utils.adapter import adapter_config


class Actor(Base, ElasticMixin):
    """Simple movie extension mapping"""

    __tablename__ = 'actors'

    id = Column(types.String(40), primary_key=True)
    name = Column(types.Unicode(40))
    birth_date = Column(types.Date())

    @classmethod
    def elastic_mapping(cls):
        return ESMapping(
            properties=ESMapping(
                actor=ESObject(
                    'actor',
                    properties=ESMapping(
                        ESText('name'),
                        ESField('birth_date')
                    )
                )
            )
        )


class TestMapping(TestCase):

    def test_base_mapping(self):
        mapping = Movie.elastic_mapping()
        self.assertIn('document_type', mapping.properties)
        self.assertEqual(mapping.properties['document_type'].parts['type'], 'keyword')

    def test_mapping_update(self):
        movie_mapping = Movie.elastic_mapping()
        actor_mapping = Actor.elastic_mapping()
        movie_mapping.update(actor_mapping)
        self.assertIn('document_type', movie_mapping.properties)
        self.assertIn('actor', movie_mapping.properties)
        self.assertIn('birth_date', movie_mapping.properties['actor'].properties)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_dotdict module

This module is used to test DotDict class.
"""

__docformat__ = 'restructuredtext'

from unittest import TestCase

from pyams_elastic.docdict import DotDict


class TestDotDict(TestCase):
    """DotDict class test case"""

    def test_empty(self):
        """Test empty dict"""
        dd = DotDict()  # pylint: disable=invalid-name
        self.assertEqual(len(dd), 0)

    def test_get(self):
        """Test getter"""
        dd = DotDict({'a': 42,  # pylint: disable=invalid-name
                      'b': 'hello'})
        self.assertEqual(dd['b'], 'hello')
        self.assertEqual(dd.b, 'hello')

    def test_recursive(self):
        """Test recursion"""
        dd = DotDict({'a': 42,  # pylint: disable=invalid-name
                      'b': {'one': 1,
                            'two': 2,
                            'three': 3}})
        self.assertEqual(dd['b']['two'], 2)
        self.assertEqual(dd.b.two, 2)

    def test_recursive_list(self):
        """Test recursive list"""
        dd = DotDict({  # pylint: disable=invalid-name
            'organization': 'Avengers',
            'members': [
                {'id': 1, 'name': 'Bruce Banner'},
                {'id': 2, 'name': 'Tony Stark'},
                {'id': 3, 'name': 'Steve Rogers'},
                {'id': 4, 'name': 'Natasha Romanoff'}
            ]
        })
        self.assertEqual(dd.members[1].name, 'Tony Stark')

    def test_set(self):
        """Test setter"""
        dd = DotDict({'a': 4,  # pylint: disable=invalid-name
                      'b': 9})
        dd.c = 16
        self.assertEqual(dd.c, 16)
        self.assertEqual(dd['c'], 16)

    def test_del(self):
        """Test deletion"""
        dd = DotDict({'a': 123,  # pylint: disable=invalid-name
                      'b': 456})
        del dd.b
        self.assertEqual(dict(dd), {'a': 123})

    def test_repr(self):
        """Test repr"""
        dd = DotDict({'a': 1})  # pylint: disable=invalid-name
        self.assertEqual(repr(dd), "<DotDict({'a': 1})>")



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.task.interfaces module

This module defines interface of PyAMS_elastic scheduler task,
which can be used to handle Elasticsearch regular queries.
"""

from zope.interface import Interface, Invalid, invariant
from zope.schema import Object, Text, TextLine

from pyams_elastic.interfaces import IElasticClientInfo
from pyams_scheduler.interfaces import ITask
from pyams_utils.schema import TextLineListField


__docformat__ = 'restructuredtext'

from pyams_elastic import _  # pylint: disable=ungrouped-imports


class IElasticTaskInfo(Interface):
    """Elasticsearch scheduler task interface"""

    connection = Object(title=_("Elasticsearch connection"),
                        schema=IElasticClientInfo,
                        required=True)

    query = Text(title=_("Query"),
                 description=_("Complete Elasticsearch query, in JSON format"),
                 required=True)

    expected_results = TextLine(title=_("Expected results count"),
                                description=_("Number of expected results; you can enter a "
                                              "single number, or a range by entering two "
                                              "numbers separated by a dash; an error status "
                                              "will be returned if the number of results is "
                                              "not in the given range; if the input is left "
                                              "empty, all queries will return an error"),
                                required=False)

    @invariant
    def check_expected_results(self):
        """Check format of expected results entry"""
        expected = self.expected_results
        if expected:
            try:
                if '-' in expected:  # pylint: disable=unsupported-membership-test
                    mini, maxi = map(int, expected.split('-'))  # pylint: disable=no-member
                    if mini > maxi:
                        raise ValueError("Minimum value must be lower or equal to maximum value")
                else:
                    _value = int(expected)
            except ValueError as exc:
                raise Invalid(_("Expected results must be a single positive number, or two "
                                "positive numbers separated by a dash")) from exc

    log_fields = TextLineListField(title=_("Log output fields"),
                                   description=_("List of results fields to include in task "
                                                 "log output report"),
                                   required=False)


class IElasticTask(ITask, IElasticTaskInfo):
    """Elasticsearch task interface"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.mixin module

This module provides mixin class which can be used to define classes of documents
which can be easily indexed and searched.
"""

import copy

__docformat__ = 'restructuredtext'


class ElasticMixin:
    """
    Mixin for SQLAlchemy classes that use ESMapping.
    """

    @classmethod
    def elastic_mapping(cls):
        """
        Return an ES mapping for the current class. Should basically be some
        form of ``return ESMapping(...)``.
        """
        raise NotImplementedError("ES classes must define a mapping")

    def elastic_document(self):
        """Apply the class ES mapping to the current instance."""
        return self.elastic_mapping()(self)

    @classmethod
    def get_document_type(cls):
        """
        Return a string keyword matching document type
        """
        return cls.__name__

    @property
    def document_type(self):
        """Document type getter"""
        return self.__class__.get_document_type()


class ESMapping:
    """
    ESMapping defines a tree-like DSL for building Elastic Search mappings.

    Calling dict(es_mapping_object) produces an Elastic Search mapping
    definition appropriate for pyes.

    Applying an ESMapping to another object returns an Elastic Search document.
    """

    def __init__(self, *args, **kwargs):
        self.filter = kwargs.pop("filter", None)
        self.name = kwargs.pop("name", None)
        self.attr = kwargs.pop("attr", None)

        # Automatically map the id field
        self.parts = {
            "_id": ESField("_id", attr="id")
        }

        # Map implicit args
        for arg in args:
            self.parts[arg.name] = arg

        # Map explicit kwargs
        for key, value in kwargs.items():  # pylint: disable=invalid-name
            if isinstance(value, dict):
                value, value.parts = ESMapping(), value  # pylint: disable=invalid-name
            if isinstance(value, ESMapping):
                value.name = key
            self.parts[key] = value

    def __iter__(self):
        for key, value in self.parts.items():  # pylint: disable=invalid-name
            if isinstance(value, ESMapping):
                value = dict(value)  # pylint: disable=invalid-name
            if value:
                yield key, value

    iteritems = __iter__
    items = __iter__

    def __contains__(self, key):
        return key in self.parts

    def __getitem__(self, key):
        return self.parts[key]

    def __setitem__(self, key, value):  # pylint: disable=invalid-name
        self.parts[key] = value

    def update(self, mapping):  # pylint: disable=invalid-name
        """
        Return a copy of the current mapping merged with the properties of
        another mapping. *update* merges just one level of hierarchy and uses
        simple assignment below that.
        """
        def is_mapping(src):
            return hasattr(src, "parts") and src.parts

        def merge_once(map1, map2):
            for key, value in map2.parts.items():
                if key in map1 and is_mapping(map1[key]) and is_mapping(value):
                    map1[key].parts.update(value.parts)
                else:
                    map1[key] = value
            return map1

        return merge_once(copy.copy(self), copy.copy(mapping))

    @property
    def properties(self):
        """
        Return the dictionary {name: property, ...} describing the top-level
        properties in this mapping, or None if this mapping is a leaf.
        """
        props = self.parts.get("properties")
        if props:
            return props.parts
        return None

    def __call__(self, instance):
        """
        Apply this mapping to an instance to return a document.

        Returns a dictionary {name: value, ...}.
        """
        if self.attr or self.name:
            instance = getattr(instance, self.attr or self.name)
        if self.filter:
            instance = self.filter(instance)
        if self.properties is None:
            return instance
        return dict((key, value(instance)) for key, value in self.properties.items())


class ESProp(ESMapping):
    """A leaf property."""

    # pylint: disable=redefined-builtin,super-init-not-called
    def __init__(self, name, filter=None, attr=None, **kwargs):
        self.name = name
        self.attr = attr
        self.filter = filter
        self.parts = kwargs


class ESField(ESProp):
    """
    A leaf property that doesn't emit a mapping definition.

    This behavior is useful if you want to allow Elastic Search to
    automatically construct an appropriate mapping while indexing.
    """

    def __iter__(self):
        return iter(())


class ESText(ESProp):
    """A string text property"""

    def __init__(self, name, **kwargs):
        ESProp.__init__(self, name, type="text", **kwargs)


class ESKeyword(ESProp):
    """A keyword text property"""

    def __init__(self, name, **kwargs):
        ESProp.__init__(self, name, type="keyword", **kwargs)


class ESObject(ESProp):
    """An object property"""

    def __init__(self, name, **kwargs):
        ESProp.__init__(self, name, type="object", **kwargs)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_functional module

This module provides tests cases for Elasticsearch client.
"""

__docformat__ = 'restructuredtext'

import unittest
from unittest import TestCase

from elasticsearch import NotFoundError
from elasticsearch_dsl.query import MatchAll

from pyams_elastic.client import ElasticClient
from pyams_elastic.tests.data import Base, Genre, Movie, get_data


class ClientTestLayer:
    """Base test layer"""

    @classmethod
    def setUp(cls):
        cls.client = ElasticClient(servers=['elasticsearch:9200'],
                                   index='pyams_elastic_tests',
                                   use_transaction=False)
        if cls.client.es.indices.exists(cls.client.index):
            cls.client.delete_index()

    @classmethod
    def tearDown(cls):
        cls.client.delete_index()
        cls.client.close()


class TestClient(TestCase):
    """Elasticsearch client test"""

    layer = ClientTestLayer

    @classmethod
    def setUpClass(cls):
        cls.layer.setUp()

    @classmethod
    def tearDownClass(cls):
        cls.layer.tearDown()

    @property
    def client(self):
        return self.layer.client

    def test_ensure_index(self):
        """Test for index creation"""
        # First ensure it with no args.
        self.client.ensure_index()
        # Recreate.
        self.client.ensure_index(recreate=True)
        # Delete explicitly.
        self.client.delete_index()
        # One more time.
        self.client.ensure_index(recreate=True)

    def test_ensure_mapping_recreate(self):
        """Test mapping creation"""
        # First create.
        self.client.ensure_mapping(Movie)
        # Recreate.
        self.client.ensure_mapping(Movie, recreate=True)

    def test_ensure_all_mappings(self):
        """Test all mappings creation"""
        self.client.ensure_index(recreate=True)
        self.client.ensure_all_mappings(Base)

    def test_get_mappings(self):
        """Test get mappings"""
        mapping = self.client.get_mappings()
        self.assertEqual(mapping['properties']['title'],
                         {'type': 'text'})
        self.assertIn('director', mapping['properties'])

    def test_disable_indexing(self):
        """Test disabled indexing"""
        self.client.ensure_index(recreate=True)

        self.client.disable_indexing = True
        genre = Genre(title='Procedural Electronica')
        self.client.index_object(genre)

        num = self.client.query(Genre, query='Electronica').count()
        self.assertEqual(num, 0)

        self.client.delete_object(genre)
        self.client.disable_indexing = False

    def test_index_and_delete_document(self):
        """Test index and delete document"""
        doc = dict(document_type='Answer',
                   question='What is the ultimate question?')
        doc_id = 42

        self.client.index_document(id=doc_id,
                                   doc=doc)
        self.client.refresh()

        query = self.client.query().ids(doc_id)
        self.assertEqual(query.count(), 1)

        self.client.delete_document(id=doc_id)
        self.client.refresh()

        query = self.client.query().ids(doc_id)
        self.assertEqual(query.count(), 0)

    def test_index_and_delete_object(self):
        genre = Genre(title='Sci-Fi Romance')
        self.client.index_object(genre)
        self.client.refresh()

        query = self.client.query(Genre).match('title', 'Sci-Fi Romance')
        self.assertEqual(query.count(), 1)

        self.client.delete_object(genre)
        self.client.refresh()

        query = self.client.query(Genre).match('title', 'Sci-Fi Romance')
        self.assertEqual(query.count(), 0)

    def test_delete_nonexistent_document(self):
        with self.assertRaises(NotFoundError):
            self.client.delete_document(id=1337)

    def test_delete_nonexistent_document_safe(self):
        self.client.delete_document(id=888,
                                    safe=True)

    def test_delete_nonexistent_object(self):
        genre = Genre(title='Geriatric Philosophy')
        with self.assertRaises(NotFoundError):
            self.client.delete_object(genre)

    def test_delete_nonexistent_object_safe(self):
        genre = Genre(title='Geriatric Philosophy')
        self.client.delete_object(genre, safe=True)


class QueryTestLayer:

    @classmethod
    def setUp(cls):
        cls.client = ElasticClient(servers=['elasticsearch:9200'],
                                   index='pyams_elastic_tests',
                                   use_transaction=False)
        cls.client.ensure_index(recreate=True)
        cls.client.ensure_all_mappings(Base, recreate=True)

        cls.genres, cls.movies = get_data()

        cls.client.index_objects(cls.genres)
        cls.client.index_objects(cls.movies)
        cls.client.refresh()

    @classmethod
    def tearDown(cls):
        cls.client.delete_index()
        cls.client.close()


class TestQuery(TestCase):

    layer = QueryTestLayer

    @classmethod
    def setUpClass(cls):
        cls.layer.setUp()

    @classmethod
    def tearDownClass(cls):
        cls.layer.tearDown()

    @property
    def client(self):
        return self.layer.client

    def test_match_all(self):
        q = self.client.query(Movie)
        result = q.match_all_query().execute()
        self.assertEqual(result.total.value, 8)

    def test_query_all(self):
        result = self.client.query(Movie).execute()
        self.assertEqual(result.total.value, 8)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertIn('Metropolis', titles)

    def test_text_query(self):
        q = self.client.query(Movie)
        result = q.text_query('allen').execute()
        self.assertEqual(result.total.value, 2)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertSequenceEqual(titles, ['Annie Hall', 'Sleeper'])

    def test_text_query_with_fields(self):
        q = self.client.query(Movie)
        result = q.text_query('curtiz', fields=['director']).execute()
        self.assertEqual(result.total.value, 1)

        q = self.client.query(Movie)
        result = q.text_query('curtiz', fields=['title', 'director']).execute()
        self.assertEqual(result.total.value, 1)

        q = self.client.query(Movie)
        result = q.text_query('curtiz', fields=['title']).execute()
        self.assertEqual(result.total.value, 0)

    def test_sorted(self):
        result = self.client.query(Movie).order_by('year', desc=True).execute()
        self.assertEqual(result.total.value, 8)

        records = list(result)
        self.assertEqual(records[0].title, 'Annie Hall')
        self.assertEqual(records[0].meta.score, None)

    def test_keyword(self):
        q = self.client.query(Movie, query='hitchcock')
        result = q.execute()
        self.assertEqual(result.total.value, 3)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertIn('To Catch a Thief', titles)

    def test_filter_year_lower(self):
        q = self.client.query(Movie)
        # Movies made after 1960.
        q = q.filter_value_lower('year', 1960)
        result = q.execute()
        self.assertEqual(result.total.value, 2)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertEqual(['Annie Hall', 'Sleeper'], sorted(titles))

    def test_filter_rating_upper(self):
        q = self.client.query(Movie)
        q = q.filter_value_upper('rating', 7.5)
        result = q.execute()
        self.assertEqual(result.total.value, 3)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertIn('Destination Tokyo', titles)

    def test_filter_term_int(self):
        q = self.client.query(Movie).\
            filter_term('year', 1927)
        result = q.execute()
        self.assertEqual(result.total.value, 1)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertIn('Metropolis', titles)

    def test_filter_terms_int(self):
        q = self.client.query(Movie).\
            filter_terms('year', [1927, 1958])
        result = q.execute()
        self.assertEqual(result.total.value, 2)

        records = list(result)
        titles = set(rec.title for rec in records)
        self.assertEqual({'Metropolis', 'Vertigo'}, titles)

    def test_offset(self):
        q = self.client.query(Movie).order_by('year').offset(4)
        result = q.execute()
        self.assertEqual(result.total.value, 8)

        records = list(result)
        self.assertEqual(len(records), 4)
        self.assertEqual(records[0].title, 'Vertigo')

    def test_offset_with_start(self):
        # If you apply .execute(start=N) on a query that already has limit M,
        # the 'start position' actually used should be M+N.
        q = self.client.query(Movie).order_by('year').offset(2)
        result = q.execute(start=2)
        # XXX How should this behave?
        self.assertEqual(result.total.value, 8)

        records = list(result)
        self.assertEqual(len(records), 4)
        self.assertEqual(records[0].title, 'Vertigo')

    def test_offset_twice(self):
        q = self.client.query(Movie).order_by('year').offset(4)
        with self.assertRaises(ValueError):
            q.offset(7)

    def test_limit(self):
        q = self.client.query(Movie).order_by('year').limit(3)
        result = q.execute()
        self.assertEqual(result.total.value, 8)

        records = list(result)
        self.assertEqual(len(records), 3)
        self.assertEqual(records[0].title, 'Metropolis')

    def test_limit_with_size(self):
        q = self.client.query(Movie).order_by('year').limit(6)
        result = q.execute(size=3)
        # XXX How should this behave?
        self.assertEqual(result.total.value, 8)

        records = list(result)
        self.assertEqual(len(records), 3)
        self.assertEqual(records[0].title, 'Metropolis')

    def test_limit_twice(self):
        q = self.client.query(Movie).order_by('year').limit(3)
        with self.assertRaises(ValueError):
            q.limit(5)

    def test_count(self):
        q = self.client.query(Movie)
        self.assertEqual(q.count(), 8)

    def test_get_tuple(self):
        genre = Genre(title='Mystery')
        record = self.client.get(('Genre', genre.id))
        self.assertEqual(record.title, 'Mystery')

    def test_get_object(self):
        genre = Genre(title='Mystery')
        record = self.client.get(genre)
        self.assertEqual(record.title, 'Mystery')

    def test_add_range_aggregate(self):
        q = self.client.query(Movie). \
            add_range_aggregate(name='era_hist',
                                field='year',
                                ranges=[
                                    {"to": 1950},
                                    {"from": 1950, "to": 1970},
                                    {"from": 1970, "to": 1990},
                                    {"from": 1990}
                                ])

        result = q.execute()
        aggregations = result.aggregations
        self.assertEqual('era_hist' in aggregations, True)
        histogram = aggregations['era_hist']

        self.assertEqual(len(histogram), 4)
        self.assertEqual(histogram[1]['from'], 1950)
        self.assertEqual(histogram[1]['to'], 1970)
        self.assertEqual(histogram[1]['doc_count'], 3)

    def test_add_term_aggregate(self):
        q = self.client.query(Movie).\
            add_term_aggregate(name='genre_hist',
                               field='genre_title',
                               size=3)

        result = q.execute()
        aggregations = result.aggregations
        self.assertEqual('genre_hist' in aggregations, True)
        histogram = aggregations['genre_hist']

        self.assertEqual(len(histogram), 3)
        self.assertEqual(histogram[0]['doc_count'], 3)
        self.assertEqual(histogram[0]['key'], 'Mystery')

    def test_raw_query(self):
        raw_query = MatchAll()
        q = self.client.query(Movie, query=raw_query)
        result = q.execute()
        self.assertEqual(result.total.value, 8)

    def test_query_fields(self):
        q = self.client.query(Movie, query='hitchcock')
        result = q.execute(fields=['title'])
        self.assertEqual(result.total.value, 3)

        records = list(result)
        titles = [rec.title for rec in records]
        self.assertIn('To Catch a Thief', titles)

    def test_add_term_suggester(self):
        q = self.client.query(Movie).\
            add_term_suggester('suggest1',
                               field='title',
                               text='vrtigo')

        result = q.execute()
        self.assertIn('suggest1', result.suggests)
        suggest = result.suggests['suggest1']
        options = suggest[0].options
        self.assertGreater(len(options), 0)
        self.assertEqual(options[0].text, 'vertigo')


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.dotdict module

This module defines a utility class which is used as a dict, but which also allows
dotted access to keys.

When instantiated from a source dict, the source will be crawled recursively to convert
all sub-dicts to this class.
"""

__docformat__ = 'restructuredtext'


class DotDict(dict):
    """A utility class which behaves like a dict, but also allows dot-access to keys"""

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, d=None):  # pylint: disable=super-init-not-called
        if d is None:
            d = {}
        for key, value in d.items():
            if hasattr(value, 'keys'):
                value = DotDict(value)
            if isinstance(value, list):
                value = [DotDict(el) if hasattr(el, 'keys') else el
                         for el in value]
            self[key] = value

    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, dict.__repr__(self))



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.test_task module

This module provides unit tests for PyAMS scheduler task for Elasticsearch.
"""

__docformat__ = 'restructuredtext'

from io import StringIO
from unittest import TestCase

from zope.interface import Invalid
from zope.schema._bootstrapinterfaces import WrongType

from pyams_elastic.client import ElasticClient, ElasticClientInfo
from pyams_elastic.task import ElasticTask
from pyams_elastic.task.interfaces import IElasticTaskInfo
from pyams_elastic.tests.data import Base, get_data
from pyams_scheduler.interfaces.task import TASK_STATUS_ERROR, TASK_STATUS_OK


class TestElasticTask(TestCase):

    def test_schema(self):
        task = ElasticTask()
        IElasticTaskInfo.validateInvariants(task)

    def test_schema_with_zero_value(self):
        task = ElasticTask()
        with self.assertRaises(WrongType):
            task.expected_results = 0

    def test_schema_with_wrong_value(self):
        task = ElasticTask()
        with self.assertRaises(WrongType):
            task.expected_results = -10

    def test_schema_with_int_value(self):
        task = ElasticTask()
        task.expected_results = '10'
        IElasticTaskInfo.validateInvariants(task)

    def test_schema_with_negative_value(self):
        task = ElasticTask()
        task.expected_results = '-10'
        with self.assertRaises(Invalid):
            IElasticTaskInfo.validateInvariants(task)

    def test_schema_with_range(self):
        task = ElasticTask()
        task.expected_results = '0-10'
        IElasticTaskInfo.validateInvariants(task)

    def test_schema_with_bad_range(self):
        task = ElasticTask()
        task.expected_results = '10-0'
        with self.assertRaises(Invalid):
            IElasticTaskInfo.validateInvariants(task)

    def test_task(self):
        client = ElasticClient(servers=['elasticsearch:9200'],
                               index='pyams_elastic_tests',
                               use_transaction=False)
        client.ensure_index(recreate=True)
        client.ensure_all_mappings(Base)
        genres, movies = get_data()
        client.index_objects(genres)
        client.index_objects(movies)
        client.refresh()

        task_info = ElasticClientInfo()
        task_info.servers = ['elasticsearch:9200']
        task_info.index = 'pyams_elastic_tests'
        task = ElasticTask()
        task.connection = task_info
        task.query = '''{
            "query": {
                "bool": {
                    "filter": {
                        "term": {
                            "year": 1977
                        }
                    }
                }
            },
            "size": 10,
            "_source": [
                "title",
                "document_type"
            ]
        }'''
        task.expected_results = '8'
        task.log_fields = ['title', 'missing.field']

        # check for error
        report = StringIO()
        status, results = task.run(report)
        self.assertEqual(status, TASK_STATUS_ERROR)

        report.seek(0)
        output = report.getvalue()
        self.assertIn("expected results: 8", output)
        self.assertIn("query results: 1", output)
        report.close()

        # check for undefined results
        task.expected_results = None
        task.log_fields = ['title', 'missing.field']
        report = StringIO()
        status, results = task.run(report)
        self.assertEqual(status, TASK_STATUS_ERROR)

        report.seek(0)
        output = report.getvalue()
        self.assertIn("expected results: --", output)
        self.assertIn("query results: 1", output)
        report.close()

        # check for OK
        task.expected_results = '1'
        task.log_fields = ['title', 'missing.field']
        report = StringIO()
        status, results = task.run(report)
        self.assertEqual(status, TASK_STATUS_OK)

        report.seek(0)
        output = report.getvalue()
        self.assertIn("expected results: 1", output)
        self.assertIn("query results: 1", output)
        report.close()

        # check for range
        task.expected_results = '0-1'
        task.log_fields = ['title', 'missing.other']
        report = StringIO()
        status, results = task.run(report)
        self.assertEqual(status, TASK_STATUS_OK)

        report.seek(0)
        output = report.getvalue()
        self.assertIn("expected results: 0-1", output)
        self.assertIn("query results: 1", output)
        report.close()

        client.delete_index()
        client.close()



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.task module

This module defines a PyAMS_scheduler task which can be used to schedule
Elasticsearch queries, and send notifications on (un)expected values.
"""

import sys
import traceback

from elasticsearch import ElasticsearchException
from zope.schema.fieldproperty import FieldProperty

from pyams_elastic.client import ElasticClient
from pyams_elastic.docdict import DotDict
from pyams_elastic.task.interfaces import IElasticTask
from pyams_scheduler.interfaces.task import TASK_STATUS_ERROR, TASK_STATUS_OK
from pyams_scheduler.task import Task
from pyams_utils.factory import factory_config


__docformat__ = 'restructuredtext'

from pyams_elastic import _  # pylint: disable=ungrouped-imports


@factory_config(IElasticTask)
class ElasticTask(Task):
    """Elasticsearch task"""

    label = _("Elasticsearch query")
    icon_class = 'fab fa-searchengin'

    connection = FieldProperty(IElasticTask['connection'])
    query = FieldProperty(IElasticTask['query'])
    expected_results = FieldProperty(IElasticTask['expected_results'])
    log_fields = FieldProperty(IElasticTask['log_fields'])

    def run(self, report, **kwargs):  # pylint: disable=unused-argument,too-many-locals,too-many-branches
        """Run Elasticsearch query task"""
        try:  # pylint: disable=too-many-nested-blocks
            client = ElasticClient(using=self.connection,
                                   use_transaction=False)
            try:
                report.write('Elasticsearch query output\n'
                             '==========================\n')
                results = client.es.search(self.query,
                                           index=self.connection.index)
                hits = DotDict(results['hits'])
                expected = self.expected_results
                total = hits.total
                if isinstance(total, DotDict):
                    total = total.value
                report.write(' - expected results: {}\n'.format(expected or '--'))
                report.write(' - total results: {}\n'.format(total))
                report.write(' - query results: {}\n'.format(len(hits.hits)))
                report.write('==========================\n')
                if expected:
                    if '-' in expected:
                        mini, maxi = map(int, expected.split('-'))
                    else:
                        mini = maxi = int(expected)
                    if not mini <= total <= maxi:
                        if self.log_fields:
                            for hit in hits.hits:
                                result = hit['_source']
                                for field in self.log_fields:
                                    record = result
                                    try:
                                        for attribute in field.split('.'):
                                            record = record[attribute]
                                        report.write(' - {}: {}\n'.format(field, record))
                                    except KeyError:
                                        report.write(' - {}: no value\n'.format(field))
                                report.write('==========================\n')
                        return TASK_STATUS_ERROR, results
                    return TASK_STATUS_OK, results
                return TASK_STATUS_ERROR, results
            finally:
                client.close()
        except ElasticsearchException:
            etype, value, tb = sys.exc_info()  # pylint: disable=invalid-name
            report.write('\n\n'
                         'An Elasticsearch error occurred\n'
                         '===============================\n')
            report.write(''.join(traceback.format_exception(etype, value, tb)))
            return TASK_STATUS_ERROR, None



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_elastic.tests.data module

This module is used to create test data.
"""

__docformat__ = 'restructuredtext'

from hashlib import sha1

from sqlalchemy import Column, ForeignKey, orm, types
from sqlalchemy.ext.declarative import declarative_base
from zope.interface import provider

from pyams_elastic.interfaces import IElasticMapping
from pyams_elastic.mixin import ESField, ESKeyword, ESMapping, ESText, ElasticMixin


Base = declarative_base()


@provider(IElasticMapping)
class Genre(Base, ElasticMixin):

    __tablename__ = 'genres'

    id = Column(types.String(40), primary_key=True)
    title = Column(types.Unicode(40))

    def __init__(self, *args, **kwargs):
        Base.__init__(self, *args, **kwargs)
        self.id = sha1(self.title.encode('utf-8')).hexdigest()

    @classmethod
    def elastic_mapping(cls):
        return ESMapping(
            # analyzer='content',
            properties=ESMapping(
                ESKeyword('document_type'),
                ESText('title')))


@provider(IElasticMapping)
class Movie(Base, ElasticMixin):

    __tablename__ = 'movies'

    id = Column(types.String(40), primary_key=True)
    title = Column(types.Unicode(40))
    director = Column(types.Unicode(40))
    year = Column(types.Integer)
    rating = Column(types.Numeric)
    genre_id = Column(None, ForeignKey('genres.id'))

    genre = orm.relationship('Genre')

    def __init__(self, *args, **kwargs):
        Base.__init__(self, *args, **kwargs)
        self.id = sha1(self.title.encode('utf-8')).hexdigest()

    @property
    def genre_title(self):
        return self.genre and self.genre.title or ''

    @classmethod
    def elastic_mapping(cls):
        return ESMapping(
            properties=ESMapping(
                ESKeyword('document_type'),
                ESText('title'),
                ESText('director'),
                ESField('year'),
                ESField('rating'),
                ESKeyword('genre_title')))


class Unindexed(Base):
    # Does not inherit from ElasticMixin.

    __tablename__ = 'unindexed'

    id = Column(types.Integer, primary_key=True)


def get_data():
    mystery = Genre(title='Mystery')
    comedy = Genre(title='Comedy')
    action = Genre(title='Action')
    drama = Genre(title='Drama')

    genres = [mystery, comedy, action, drama]

    movies = [
        Movie(
            title='To Catch a Thief',
            director='Alfred Hitchcock',
            year=1955,
            rating=7.5,
            genre=mystery,
            genre_id=mystery.id,
        ),
        Movie(
            title='Vertigo',
            director='Alfred Hitchcock',
            year=1958,
            rating=8.5,
            genre=mystery,
            genre_id=mystery.id,
        ),
        Movie(
            title='North by Northwest',
            director='Alfred Hitchcock',
            year=1959,
            rating=8.5,
            genre=mystery,
            genre_id=mystery.id,
        ),
        Movie(
            title='Destination Tokyo',
            director='Delmer Daves',
            year=1943,
            rating=7.1,
            genre=action,
            genre_id=action.id,
        ),
        Movie(
            title='Annie Hall',
            director='Woody Allen',
            year=1977,
            rating=8.2,
            genre=comedy,
            genre_id=comedy.id,
        ),
        Movie(
            title='Sleeper',
            director='Woody Allen',
            year=1973,
            rating=7.3,
            genre=comedy,
            genre_id=comedy.id,
        ),
        Movie(
            title='Captain Blood',
            director='Michael Curtiz',
            year=1935,
            rating=7.8,
            genre=action,
            genre_id=action.id,
        ),
        Movie(
            title='Metropolis',
            director='Fritz Lang',
            year=1927,
            rating=8.4,
            genre=drama,
            genre_id=drama.id,
        )]
    return genres, movies
